"use strict";
this.name = "StationDockControl_Interface";
this.author = "phkb";
this.copyright = "2015 phkb";
this.description = "Controls the Interfaces (F4) traffic control screen.";
this.license = "CC BY-NC-SA 3.0";

this._maxpage = 0; // total number of pages available for display
this._curpage = 1; // the current page being displayed
this._msRows = 16; // rows to display on the mission screen
this._msCols = 32; // columns to display on the mission screen
this._showAll = false; // flag that controls whether the short or long list of ships is shown
this._displayMode = 0; // controls the type of display on the Traffic Control screen (0 = dock list, 1 = individual ship)
this._selectedItem = {}; // item selected on the Traffic Control screen
this._selectedIndex = 0; // the array index of the selected item
this._sortCol = "departureTime"; // default sort column
this._sortDir = 1; // default sort direction
this._rsnInstalled = false; // indicates whether randomshipnames OXP is installed
this._playerDockingBay = ""; // stores the docking bay assigned to the player
this._playerDockTime = 0; // stores the time the player docked at the station
this._initialPageLimit = 3; // sets the initial number of pages that will be shown when the list is first opened
this._refreshPageTime = 20; // number of seconds between page refreshes, when the view is open
this._rsnInMFD = true;
this._mfdAllStations = false;

this._viewing = []; // subset of main array used for viewing docked vessels on the interface screen
this._SDC = null; // link to the main station dock control worldscript
this._debug = false; // controls output of information to the log. Slaved to StationDockControl._debug
this._selectedStation = null; // the station whose dock list is being viewed. When not in debug mode, this will always be the players docked station.
this._refresh = null; // link to a timer that refreshes the dock list
this._stationModel = false; // controls whether a model of the station is included on the station traffic contorl screen
this._shipModel = true; // controls whether a model of the selected ship is included on the ship information screen
this._spinModel = false; // controls whether the ship model will be spinning or not
this._showDataKey = false; // include the ship's datakey and personality info on the ship details page
this._showEquipment = false; // include an option to view ship equipment
this._mode = 0; // display mode: 0 = full debug, 1 = standard operations
this._itemColor = "yellowColor";
this._menuColor = "orangeColor";
this._exitColor = "yellowColor";
this._disabledColor = "darkGrayColor";
this._defaultZoomFactor = 3.6;
this._destViewingRange = 12; // the number of hours until departure time that the destination will be viewable

this._externalMenu = [];

this._sdciConfig = {
	Name: this.name,
	Display: "Interface Options",
	Alias: "Station Dock Control",
	Alive: "_sdciConfig",
	Bool: {
		B0: {
			Name: "_shipModel",
			Def: true,
			Desc: "Show ship model"
		},
		B1: {
			Name: "_spinModel",
			Def: false,
			Desc: "Spin ship model"
		},
		B2: {
			Name: "_showDataKey",
			Def: false,
			Desc: "Show data key"
		},
		B3: {
			Name: "_stationModel",
			Def: false,
			Desc: "Show station model"
		},
		B4: {
			Name: "_rsnInMFD",
			Def: true,
			Desc: "Show RSN name in MFD"
		},
		B5: {
			Name: "_mfdAllStations",
			Def: false,
			Desc: "MFD available on all stations"
		},
		Info: "0 - Shows the ship model when viewing entries.\n1 - Ship models will spin when viewed.\n2 - Displays the shipDataKey when viewing entries.\n3 - Shows the station model when viewing main list.\n4 - Shows RSN ship name in docking MFD\n5 - Launch queue MFD available at all stations"
	},
	SInt: {
		S0: {
			Name: "_mode",
			Def: 1,
			Min: 0,
			Max: 1,
			Desc: "Display mode"
		},
		Info: "SDC interface display mode. 0 = full debug, 1 = standard (default)"
	},
};

// note: if a station has the "has_npc_traffic" flag set to false, it will override whatever is in this list
// this item controls which station roles will *not* have the "Traffic Control" interface
this._noDockView = ["slaver_base", "rrs_slaverbase", "rrs_group_spacestation", "rrs_astromine", "rrs-mining-outpost", "liners_liner", "navystat",
	"casinoship", "astromine", "random_hits_any_spacebar", "comczgf", "thecollector_adck_coriolis-station", "jaguar_company_base", "GW-fuel-processor", "GW-comms-station",
	"laveAcademy_academy", "planetFall_surface", "generationship", "IST_genship carrier", "pirate-cove"
];

// default zoom is 3.6
// these values make ships look bigger or smaller, as appropriate
this._zoomOverride = {
	"Adder": 4.6,
	"Adder (transit)": 4.6,
	"Adder (civilian)": 4.6,
	"Adder (military)": 4.6,
	"Adder (performance)": 4.6,
	"Anaconda": 2.3,
	"Anaconda MMS": 2.3,
	"Anaconda (service)": 2.3,
	"Anaconda (civilian)": 2.3,
	"Anaconda (military)": 2.3,
	"Anaconda (performance)": 2.3,
	"Aphid": 3.2,
	"Arafura": 3,
	"Boa Class Cruiser": 2.5,
	"Boa Class Cruiser MMS": 2.5,
	"Boa MkII (transit)": 2.5,
	"Boa MkII (civilian)": 2.5,
	"Boa MkII (military)": 2.5,
	"Boa MkII (service)": 2.5,
	"Boa": 2.5,
	"Boa MMS": 2.5,
	"Boa MkI (service)": 2.5,
	"Boa MkI (civilian)": 2.5,
	"Boa MkI (military)": 2.5,
	"Boa MkI (performance)": 2.5,
	"Bug": 4.4,
	"Chameleon": 3.3,
	"Chopped Cobra": 3,
	"Cobra Mark I": 4.6,
	"Cobra MkI (transit)": 4.6,
	"Cobra MkI (civilian)": 4.6,
	"Cobra MkI (military)": 4.6,
	"Cobra MkI (performance)": 4.6,
	"Cobra Mark II-X": 4,
	"Cobra Mark III": 3,
	"Cobra Mark III MMS": 3,
	"Cobra MkIII (transit)": 3,
	"Cobra MkIII (civilian)": 3,
	"Cobra MkIII (military)": 3,
	"Cobra MkIII (service)": 3,
	"Cobra Mark III-XT": 3,
	"Cobra Mark IV": 2.9,
	"Cobra Mark IV MMS": 2.9,
	"Cobra Rapier": 3,
	"Copperhead Mk2": 3,
	"Cruzer": 3,
	"D.T.T. Atlas": 2.5,
	"D.T.T. Cyclops": 2.5,
	"D.T.T. Kraken": 2.8,
	"D.T.T. MK-1": 3.4,
	"D.T.T. Snake Charmer": 2.7,
	"D.T.T. War Lance": 3.3,
	"Fer-de-Lance": 3.3,
	"Fer de Lance (transit)": 3.3,
	"Fer de Lance (service)": 3.3,
	"Fer de Lance (military)": 3.3,
	"Fer de Lance (performance)": 3.3,
	"Gecko": 4.2,
	"Gecko (transit)": 4.2,
	"Gecko (service)": 4.2,
	"Gecko (military)": 4.2,
	"Gecko (performance)": 4.2,
	"Griffin Mk I": 3.3,
	"Griff Boa Prototype": 2.5,
	"Hognose": 4.3,
	"Iguana": 3.2,
	"Krait": 4.2,
	"Krait (transit)": 4.2,
	"Krait (service)": 4.2,
	"Krait (military)": 4.2,
	"Krait (performance)": 4.2,
	"Lira": 1.9,
	"Mamba Escort": 4.2,
	"Mamba": 4.2,
	"Mamba (transit)": 4.2,
	"Mamba (service)": 4.2,
	"Mamba (military)": 4.2,
	"Mamba (performance)": 4.2,
	"Miner Cobra Mark III": 3,
	"Mining Transporter": 4.6,
	"Monitor Mark II": 3,
	"Monitor": 3,
	"Moray Medical Boat": 4.2,
	"Moray (service)": 4.2,
	"Moray Star Boat": 4.2,
	"Moray (civilian)": 4.2,
	"Moray (military)": 4.2,
	"Mussurana": 3,
	"Night Adder": 4.2,
	"Ophidian": 3,
	"Orbital Shuttle": 4.6,
	"Pitviper": 3.4,
	"Pitviper (Vindicator)": 3.4,
	"Pitviper (Buccaneer)": 3.4,
	"Pitviper (Bountyhunter)": 3.4,
	"Pitviper S.E. Beast": 3.4,
	"Pitviper Mark II": 3,
	"Python Class Cruiser": 2.5,
	"Python ET Special": 2.6,
	"Python ET Special MMS": 2.6,
	"Python": 2.5,
	"Python MMS": 2.5,
	"Python (transit)": 2.5,
	"Python (civilian)": 2.5,
	"Python (military)": 2.5,
	"Python (performance)": 2.5,
	"Python SG": 2.5,
	"Python SG-MMS": 2.5,
	"Python BattleCruiser MMS": 2.5,
	"Serpent Class Cruiser": 2.5,
	"Sidewinder Scout Ship": 4.5,
	"Sidewinder (transit)": 4.5,
	"Sidewinder (service)": 4.5,
	"Sidewinder (civilian)": 4.5,
	"Sidewinder (performance)": 4.5,
	"Sidewinder Special": 4.3,
	"Skat": 3.4,
	"Taxi Cab": 4.6,
	"Tembo": 2.2,
	"Transporter": 4.6,
	"Transporter (civilian)": 4.6,
	"Transporter (service)": 4.6,
	"Transporter (military)": 4.6,
	"Transporter (performance)": 4.6,
	"Worm": 4.6,
	"Worm (transit)": 4.6,
	"Worm (service)": 4.6,
	"Worm (military)": 4.6,
	"Worm (performance)": 4.6,
	"Yasen-N": 3.2,
	"Yasen-N 'Advanced' variant 1": 3.2,
	"Yasen-N 'Advanced' variant 2": 3.2,
	"Yasen-N 'Advanced' variant 3": 3.2,
	"YellOo Cab": 4.4,
};

this._trueValues = ["yes", "1", 1, "true", true];

//=============================================================================================================
// system functions
//-------------------------------------------------------------------------------------------------------------
this.startUp = function () {
	this._hudHidden = false;
	this._dockListOpen = false;
}

//-------------------------------------------------------------------------------------------------------------
this.startUpComplete = function () {
	// register our settings, if Lib_Config is present
	if (worldScripts.Lib_Config) worldScripts.Lib_Config._registerSet(this._sdciConfig);

	if (missionVariables.StationDockControl_ShipModel) this._shipModel = (this._trueValues.indexOf(missionVariables.StationDockControl_ShipModel) >= 0 ? true : false);
	if (missionVariables.StationDockControl_SpinModel) this._spinModel = (this._trueValues.indexOf(missionVariables.StationDockControl_SpinModel) >= 0 ? true : false);
	if (missionVariables.StationDockControl_DataKey) this._showDataKey = (this._trueValues.indexOf(missionVariables.StationDockControl_ShowDataKey) >= 0 ? true : false);
	if (missionVariables.StationDockControl_StationModel) this._stationModel = (this._trueValues.indexOf(missionVariables.StationDockControl_StationModel) >= 0 ? true : false);
	if (missionVariables.StationDockControl_RSNinMFD) this._rsnInMFD = (this._trueValues.indexOf(missionVariables.StationDockControl_RSNinMFD) >= 0 ? true : false);

	this._SDC = worldScripts.StationDockControl;

	if (this._SDC._disable === true) {
		delete this.shipDied;
		delete this.guiScreenChanged;
		delete this.shipDockedWithStation;
		delete this.shipWillLaunchFromStation;
		delete this.startUpComplete;
		return;
	}

	this._debug = this._SDC._debug;

	// is randomship names installed?
	if (worldScripts["randomshipnames"]) this._rsnInstalled = true;
	// if it isn't, we will have one extra display row on the interface screen.
	if (this._rsnInstalled === false) this._msRows += 1;

	if (missionVariables.StationDockControl_PlayerDockingBay) {
		this._playerDockingBay = missionVariables.StationDockControl_PlayerDockingBay;
	}
	if (missionVariables.StationDockControl_PlayerDockTime) {
		this._playerDockTime = missionVariables.StationDockControl_PlayerDockTime;
	} else {
		// if we don't have a saved dock time, default to 5 minutes ago
		this._playerDockTime = global.clock.adjustedSeconds - (5 * 60);
	}

	// set up the interface screen, if required
	var p = player.ship;
	if (p.dockedStation) {
		if (p.dockedStation.hasNPCTraffic && this.$checkStationRoleForView(p.dockedStation)) {
			this._selectedStation = p.dockedStation;
			this.$initInterface(p.dockedStation);
		}
	}

	// if there are external menu items in the save game, restore them here
	if (missionVariables.StationDockControl_ExternalMenu) {
		this._externalMenu = JSON.parse(missionVariables.StationDockControl_ExternalMenu);
		delete missionVariables.StationDockControl_ExternalMenu;
	}
}

//-------------------------------------------------------------------------------------------------------------
this.shipDied = function (whom, why) {
	// clean up any active timers
	if (this._refresh && this._refresh.isRunning) this._refresh.stop();
}

//-------------------------------------------------------------------------------------------------------------
this.guiScreenChanged = function (to, from) {
	if (from === "GUI_SCREEN_MISSION" && this._dockListOpen) {
		this._dockListOpen = false;
		if (this._hudHidden === false && player.ship.hudHidden === true) {
			player.ship.hudHidden = false;
		}
		if (this._refresh && this._refresh.isRunning) this._refresh.stop();
	}
}

//-------------------------------------------------------------------------------------------------------------
this.shipDockedWithStation = function (station) {
	// set up the interface screen, if required
	this._playerDockTime = global.clock.adjustedSeconds;
	if (system.ID != -1 && station.hasNPCTraffic && this.$checkStationRoleForView(station)) {
		this._selectedStation = station;
		this.$initInterface(station);
	}
}

//-------------------------------------------------------------------------------------------------------------
this.shipWillLaunchFromStation = function (station) {
	this._playerDockingBay = "";
}

//-------------------------------------------------------------------------------------------------------------
this.playerWillSaveGame = function () {
	if (this._playerDockingBay !== "") {
		missionVariables.StationDockControl_PlayerDockingBay = this._playerDockingBay;
	} else {
		delete missionVariables.StationDockControl_PlayerDockingBay;
	}
	missionVariables.StationDockControl_PlayerDockTime = this._playerDockTime;

	// save any external menu items
	if (this._externalMenu.length > 0) {
		missionVariables.StationDockControl_ExternalMenu = JSON.stringify(this._externalMenu);
	}
	missionVariables.StationDockControl_ShipModel = this._shipModel;
	missionVariables.StationDockControl_SpinModel = this._spinModel;
	missionVariables.StationDockControl_ShowDataKey = this._showDataKey;
	missionVariables.StationDockControl_StationModel = this._stationModel;
	missionVariables.StationDockControl_RSNinMFD = this._rsnInMFD;
}

//=============================================================================================================
// external interfaces

//-------------------------------------------------------------------------------------------------------------
// add a custom menu to the ship display, with a callback
// obj has following params:
// 	shipName		name of ship to add menu item for (required if pilotname not specified)
// 	shipType		ship type (required if pilotname not specified)
// 	pilotName		name of pilot (optional, but required if shipname and shiptype are blank)
// 	menuText		text to display in the menu (required)
// 	worldScript		worldScript name for the callback (required)
// 	callback		worldScript function to call in the callback (required)
// 	parameter		parameter to add to the function when called (optional)
this.$externalInterface = function $externalInterface(obj) {
	if ((!obj.shipName || obj.shipName === "") && (!obj.pilotName || obj.pilotName === "")) {
		throw "Invalid settings: shipName or pilotName must be specified";
	}
	if ((!obj.shipType || obj.shipType === "") && (!obj.pilotName || obj.pilotName === "")) {
		throw "Invalid settings: shipType or pilotName must be specified";
	}
	if (!obj.menuText || obj.menuText === "") {
		throw "Invalid settings: menuText must be specified, and cannot be blank.";
	}
	if (!obj.worldScript || obj.worldScript === "") {
		throw "Invalid settings: worldScript (worldScript name) must be specified, and cannot be blank.";
	}
	if (!obj.callback || obj.callback === "") {
		throw "Invalid settings: callback (function name) must be specified, and cannot be blank.";
	}
	var stnName = "";
	var stnIndex = 0;
	if (obj.station) {
		stnIndex = this._SDC.$getStationIndex(obj.station);
		stnName = obj.name;
	}

	this._externalMenu.push({
		shipName: (obj.shipName ? obj.shipName : ""),
		shipType: (obj.shipType ? obj.shipType : ""),
		pilotName: (obj.pilotName ? obj.pilotName : ""),
		stationName: stnName,
		stationIndex: stnIndex,
		menuText: obj.menuText,
		callback: obj.callback,
		worldScript: obj.worldScript,
		parameter: (obj.hasOwnProperty("parameter") ? obj.parameter : "")
	});
}

//-------------------------------------------------------------------------------------------------------------
// removes a custom menu from the list
// obj has following params:
// 	shipName		name of ship to remove the menu item for. optional
// 	shipType		ship type. optional
// 	pilotName		name of pilot. optional.
// 	menuText		text displayed in the menu. required
// 	worldScript		worldScript name. required.
this.$removeExternalInterface = function $removeExternalInterface(obj) {
	for (var i = this._externalMenu.length - 1; i >= 0; i--) {
		if ((!obj.shipName || (this._externalMenu[i].shipName === obj.shipName)) &&
			(!obj.shipType || (this._externalMenu[i].shipType === obj.shipType)) &&
			(!obj.pilotName || (this._externalMenu[i].pilotName === obj.pilotName)) &&
			this._externalMenu[i].menuText === obj.menuText &&
			this._externalMenu[i].worldScript === obj.worldScript) {
			this._externalMenu.splice(i, 1);
		}
	}
}

//-------------------------------------------------------------------------------------------------------------
// returns true if an external menu item exists for the ship/pilot & menu/ws combination
// obj has following params:
// 	shipName		name of ship to remove the menu item for. optional
// 	shipType		ship type. optional
// 	pilotName		name of pilot. optional.
// 	menuText		text displayed in the menu. required
// 	worldScript		worldScript name. required.
this.$externalInterfaceExists = function $externalInterfaceExists(obj) {
	for (var i = this._externalMenu.length - 1; i >= 0; i--) {
		if ((!obj.shipName || (this._externalMenu[i].shipName === obj.shipName)) &&
			(!obj.shipType || (this._externalMenu[i].shipType === obj.shipType)) &&
			(!obj.pilotName || (this._externalMenu[i].pilotName === obj.pilotName)) &&
			this._externalMenu[i].menuText === obj.menuText &&
			this._externalMenu[i].worldScript === obj.worldScript) {
			return true;
		}
	}
	return false;
}

//=============================================================================================================
// screen interfaces

//-------------------------------------------------------------------------------------------------------------
// initialise the F4 screen entries
this.$initInterface = function $initInterface(station) {
	station.setInterface(this.name, {
		title: this._selectedStation.name + " " + expandDescription("[sdc_title]"),
		category: expandDescription("[sdc_category]"),
		summary: expandDescription("[sdc_summary]"),
		callback: this.$showDockedList.bind(this)
	});
	// set up interface for switching the dock view to another station in system
	if (this._mode === 0) {
		station.setInterface(this.name + "_dv", {
			title: "Select station dock view",
			category: expandDescription("[sdc_category]"),
			summary: "Switch the dock view to another station in system",
			callback: this.$selectStationList.bind(this)
		});
	}
}

//-------------------------------------------------------------------------------------------------------------
// show the station selection list
this.$selectStationList = function $selectStationList() {
	var curChoices = {};
	var stns = system.stations;
	for (var i = 0; i < stns.length; i++) {
		if (stns[i].hasNPCTraffic) {
			curChoices["01_STN~" + stns[i].displayName] = {
				text: stns[i].displayName,
				color: "yellowColor"
			};
		}
	}

	curChoices["99_EXIT"] = {
		text: "Exit",
		color: "yellowColor"
	};
	var opts = {
		screenID: "oolite-sdc-switch-map",
		title: "Select Station Traffic Control",
		allowInterrupt: true,
		exitScreen: "GUI_SCREEN_INTERFACES",
		choices: curChoices,
		initialChoicesKey: "99_EXIT",
		message: "Select a station whose traffic control you wish to monitor\n\nCurrently viewing: " + this._selectedStation.displayName +
			"\n\nNote: This function is for testing purposes only. It will not be in the official release of this OXP."
	};
	mission.runScreen(opts, this.$selectStationHandler, this);
}

//-------------------------------------------------------------------------------------------------------------
// handles UI responses from the station selection list
this.$selectStationHandler = function $selectStationHandler(choice) {
	if (choice.indexOf("01_STN") >= 0) {
		var nam = choice.substring(7);
		var stns = system.stations;
		for (var i = 0; i < stns.length; i++) {
			if (stns[i].displayName === nam) {
				this._selectedStation = stns[i];
				break;
			}
		}
		this.$initInterface(player.ship.dockedStation);
	}
	if (choice !== "99_EXIT") {
		this.$selectStationList();
	}
}

//-------------------------------------------------------------------------------------------------------------
// initialise and open the dock list
this.$showDockedList = function $showDockedList() {
	if (this._SDC.$stationIsRepopulating(this._selectedStation) === false) {
		this._SDC.$updateShipsDocking(this._selectedStation);
		this._viewing = this._SDC.$getViewingList(this._selectedStation, false);
		if (this._selectedStation === player.ship.dockedStation) {
			this.$addPlayerToArray(this._viewing);
		}
	}
	this._curpage = 0;
	this._displayMode = 0;
	this._showAll = false;
	this.$showPage();
}

//-------------------------------------------------------------------------------------------------------------
// auto-refresh of page
this.$refreshPage = function $refreshPage() {
	// launched ships should be removed from the array elsewhere, so all we need to do is refresh the page
	if (this._dockListOpen === true) {
		if (this._SDC.$stationIsRepopulating(this._selectedStation) === false) {
			this._SDC.$updateShipsDocking(this._selectedStation);
			this._viewing = this._SDC.$getViewingList(this._selectedStation, false);
			// add the player to the dock list if the docked station is the selected station
			if (this._selectedStation === player.ship.dockedStation) this.$addPlayerToArray(this._viewing);
		}
		this.$showPage();
	}
}

//-------------------------------------------------------------------------------------------------------------
// show the dock list
this.$showPage = function $showPage() {
	function compareDeparture(a, b) {
		if ((a.departureTime * 10 + a.departureSeconds) < (b.departureTime * 10 + b.departureSeconds)) return (worldScripts.StationDockControl_Interface._sortDir * -1);
		if ((a.departureTime * 10 + a.departureSeconds) > (b.departureTime * 10 + b.departureSeconds)) return (worldScripts.StationDockControl_Interface._sortDir);
		return 0;
	}

	function compareShipName(a, b) {
		if (a.shipName < b.shipName) return (worldScripts.StationDockControl_Interface._sortDir * -1);
		if (a.shipName > b.shipName) return (worldScripts.StationDockControl_Interface._sortDir);
		return 0;
	}

	function compareShipType(a, b) {
		if (a.shipType < b.shipType) return (worldScripts.StationDockControl_Interface._sortDir * -1);
		if (a.shipType > b.shipType) return (worldScripts.StationDockControl_Interface._sortDir);
		return 0;
	}

	this._hudHidden = player.ship.hudHidden;
	if (this.$isBigGuiActive() === false) {
		player.ship.hudHidden = true;
	}
	this._dockListOpen = true;

	var p = player.ship;
	var curChoices = {};
	var text = "";
	var def = "99_EXIT";

	var hackchip = worldScripts.StationDockControl_HackChip._chipInstalled;

	// stop the timer if it's running
	if (this._refresh && this._refresh.isRunning) this._refresh.stop();

	if (this._displayMode === 0) {
		// start a timer to refresh the screen
		this._refresh = new Timer(this, this.$refreshPage, this._refreshPageTime, 0);

		switch (this._sortCol) {
			case "departureTime":
				this._viewing.sort(compareDeparture);
				break;
			case "shipType":
				this._viewing.sort(compareShipType);
				break;
			case "shipName":
				this._viewing.sort(compareShipName);
				break;
		}

		// calculate the max number of pages
		this._maxpage = Math.ceil(this._viewing.length / this._msRows);
		if (this._showAll === false && this._maxpage > this._initialPageLimit) this._maxpage = this._initialPageLimit;

		// if we only have 1 page, add some extra lines to the page and hide the next/prev page options
		var extra = 0;
		if (this._maxpage <= 1) extra = 2;

		var col1 = 21;
		var col2 = 4.5;
		var col3 = 6.5;

		// build the header
		text += this.$padTextRight(" Ship", col1);
		text += this.$padTextRight(" Status", col2);
		text += this.$padTextRight(" Destination", col3);

		var disable_screen = false;

		if (this._SDC._repopulate === true || this._SDC.$stationIsRepopulating(this._selectedStation) === true) {
			// if the data is about to be changed in the next repopulate loop, don't show anything
			text += "\n" + this.$duplicate("-", 32) + expandDescription("[sdc-data-loading]") + this.$duplicate("-", 32) + "\n";
			this._maxpage = 0;
			this._curpage = 0;
			disable_screen = true;
		} else {

			var items = 0;
			var start = this._curpage * (this._msRows + extra);
			var end = (this._curpage * (this._msRows + extra)) + (this._msRows + extra);
			var shipItem = "";
			var shipRef = 0;
			for (var i = 0; i < this._viewing.length; i++) {
				items += 1;
				if ((items - 1) >= start && (items - 1) < end) {
					shipRef += 1;
					if (shipRef === (this._msRows + extra) && this._curpage === (this._initialPageLimit - 1) && this._showAll === false) {
						curChoices["01_99_EXPAND"] = {
							text: expandDescription("[showall]"),
							color: this._menuColor,
							alignment: "LEFT"
						};
					} else {
						switch (this._sortCol) {
							case "departureTime":
							case "shipType":
								shipItem = this.$padTextRight(this._viewing[i].shipType +
										((this._viewing[i].shipName && this._viewing[i].shipName.length > 1) ? ": " + this._viewing[i].shipName : ""), col1) +
									this.$padTextRight(this.$departureTimeFormat(this._viewing[i].departureTime), col2) +
									this.$padTextRight((hackchip || this.$viewDestination(this._viewing[i]) ? this.$outputDestination(this._viewing[i]) : ""), col3);
								break;
							case "shipName":
								shipItem = this.$padTextRight(((this._viewing[i].shipName && this._viewing[i].shipName.length > 1) ? this._viewing[i].shipName + " - " : "") + this._viewing[i].shipType, col1) +
									this.$padTextRight(this.$departureTimeFormat(this._viewing[i].departureTime), col2) +
									this.$padTextRight((hackchip || this.$viewDestination(this._viewing[i]) ? this.$outputDestination(this._viewing[i]) : ""), col3);
								break;
						}

						curChoices["01_" + (shipRef < 10 ? "0" : "") + shipRef + "_SHIPITEM_" + i] = {
							text: shipItem,
							color: this._itemColor,
							alignment: "LEFT"
						};
					}
				}
				if ((items - 1) >= end) break;
			}

			for (var i = 0; i < ((this._msRows + extra) + 1) - shipRef; i++) {
				curChoices["02_SPACER_" + i.toString()] = "";
			}
		}

		if (this._maxpage > 1) {
			if (this._curpage < this._maxpage - 1) {
				curChoices["10_GOTONEXT"] = {
					text: "[option_nextpage]",
					color: this._menuColor
				};
			} else {
				curChoices["10_GOTONEXT"] = {
					text: "[option_nextpage]",
					color: this._disabledColor,
					unselectable: true
				};
			}
			if (this._curpage > 0) {
				curChoices["11_GOTOPREV"] = {
					text: "[option_prevpage]",
					color: this._menuColor
				};
			} else {
				curChoices["11_GOTOPREV"] = {
					text: "[option_prevpage]",
					color: this._disabledColor,
					unselectable: true
				};
			}
		}
		if (this._sortCol === "shipType" || disable_screen === true) {
			curChoices["30_SORTBY_SHIPTYPE"] = {
				text: "[option_sortshiptype]",
				color: this._disabledColor,
				unselectable: true
			};
		} else {
			curChoices["30_SORTBY_SHIPTYPE"] = {
				text: "[option_sortshiptype]",
				color: this._menuColor
			};
		}
		if (this._rsnInstalled === true) {
			if (this._sortCol === "shipName" || disable_screen === true) {
				curChoices["31_SORTBY_SHIPNAME"] = {
					text: "[option_sortshipname]",
					color: this._disabledColor,
					unselectable: true
				};
			} else {
				curChoices["31_SORTBY_SHIPNAME"] = {
					text: "[option_sortshipname]",
					color: this._menuColor
				};
			}
		}
		if (this._sortCol === "departureTime" || disable_screen === true) {
			curChoices["32_SORTBY_DEPART"] = {
				text: "[option_sortstatus]",
				color: this._disabledColor,
				unselectable: true
			};
		} else {
			curChoices["32_SORTBY_DEPART"] = {
				text: "[option_sortstatus]",
				color: this._menuColor
			};
		}

		if (this._sortDir === 1 || disable_screen === true) {
			curChoices["35_SORT_ASC"] = {
				text: "[option_sortasc]",
				color: this._disabledColor,
				unselectable: true
			};
		} else {
			curChoices["35_SORT_ASC"] = {
				text: "[option_sortasc]",
				color: this._menuColor
			};
		}
		if (this._sortDir === -1 || disable_screen === true) {
			curChoices["36_SORT_DESC"] = {
				text: "[option_sortdesc]",
				color: this._disabledColor,
				unselectable: true
			};
		} else {
			curChoices["36_SORT_DESC"] = {
				text: "[option_sortdesc]",
				color: this._menuColor
			};
		}
		curChoices["99_EXIT"] = {
			text: "[option_exit]",
			color: this._exitColor
		};

		var pgtitle = this._selectedStation.name + " " + expandDescription("[sdc_title_proper]") + (disable_screen === true || this._maxpage <= 1 ? "" : " - Page " + (this._curpage + 1) + " of " + ((this._maxpage > 3 && this._showAll === false) ? this._initialPageLimit : this._maxpage));
		if (defaultFont.measureString(pgtitle) > 17) {
			pgtitle = expandDescription("[sdc_title_proper]") + (disable_screen === true || this._maxpage <= 1 ? "" : " - Page " + (this._curpage + 1) + " of " + ((this._maxpage > 3 && this._showAll === false) ? this._initialPageLimit : this._maxpage));
		}
		var opts = {
			screenID: "oolite-sdc-main-map",
			title: pgtitle,
			allowInterrupt: true,
			exitScreen: "GUI_SCREEN_INTERFACES",
			choices: curChoices,
			initialChoicesKey: this._lastchoice ? this._lastchoice : def,
			message: text
		};
		if (this._stationModel === true) {
			opts["model"] = "[" + this._selectedStation.dataKey + "]";
			opts["modelPersonality"] = this._selectedStation.entityPersonality;
			opts["spinModel"] = this._spinModel;
		} else {
			opts["overlay"] = {
				name: "sdc-station.png",
				height: 546
			};
		}
	}

	// displaying details of an individual ship
	if (this._displayMode === 1) {
		var depart = this.$departureTimeFormat(this._selectedItem.departureTime);
		var column = 11;

		// ship name/type
		text += ((this._selectedItem.shipName && this._selectedItem.shipName.length > 1) ? this.$padTextRight(expandDescription("[shipdetails-shipname]"), column) + this._selectedItem.shipName + "\n" : "") +
			this.$padTextRight(expandDescription("[shipdetails-shiptype]"), column) + this._selectedItem.shipType + ((this._mode === 0) ? " (" + this._selectedItem.primaryRole + ")" : "") + "\n";

		if (this._showDataKey) {
			text += this.$padTextRight(expandDescription("[shipdetails-datakey]"), column) + this._selectedItem.shipDataKey + " (" + this._selectedItem.personality + ")\n";
			//log(this.name, "player ent person = " + player.ship.entityPersonality);
		}

		// legal status
		if (p.equipmentStatus("EQ_SCANNER_SHOW_MISSILE_TARGET") === "EQUIPMENT_OK") {
			text += this.$padTextRight(expandDescription("[shipdetails-legal]"), column) +
				(this._selectedItem.bounty === 0 ? "Clean" : (this._selectedItem.bounty > 0 && this._selectedItem.bounty <= 50 ? "Offender" : "Fugitive")) + "\n";
		}
		// docking bay/dock time
		if (depart !== "Inbound") {
			if (this._selectedItem.dockingBay !== "") {
				var pref = this.$getDockingBayPrefix(p.dockedStation);
				text += this.$padTextRight(expandDescription("[shipdetails-dockingbay]"), column) + pref + this._selectedItem.dockingBay + "\n";
			}
			text += this.$padTextRight(expandDescription("[shipdetails-docktime]"), column) + global.clock.clockStringForTime(this._selectedItem.dockTime) + "\n";
		}

		// departure time/status
		var lt = (this._selectedItem.departureTime * 60 + this._selectedItem.lastChange);
		text += this.$padTextRight(((lt > (global.clock.adjustedSeconds + 60) &&
			lt < (global.clock.adjustedSeconds + (12 * 60 * 60))) ? expandDescription("[shipdetails-departtime]") : expandDescription("[shipdetails-status]")), column) + depart + "\n";

		// destination
		switch (depart) {
			// for these items hide the destination - the ship hasn't loaded yet, so no decision on destination has been made
			case "Inbound":
			case "Docking":
			case "Unloading":
			case "Docked":
				break;
			default:
				if (hackchip || this.$viewDestination(this._selectedItem) === true) {
					text += this.$padTextRight(expandDescription("[shipdetails-destinationsystem]"), column) + System.systemNameForID(this._selectedItem.destinationSystem) + "\n";
				}
				if (this._selectedItem.destination && this._selectedItem.destination !== "") {
					text += this.$padTextRight(expandDescription("[shipdetails-destination]"), column) + this.$viewLocation(this._selectedItem.destination) + "\n";
				}
				break;
		}

		// ship groups
		if (this._selectedItem.groupName !== "" && depart !== "Inbound") {
			text += "\n";
			var group = this.$getGroupMembers(this._selectedItem.groupName, this._selectedItem);
			if (group.length > 0) {
				if (this._selectedItem.groupLeader === true) {
					text += this.$padTextRight(expandDescription("[shipdetails-groupleader]"), column) + "Yes\n";
				} else {
					text += this.$padTextRight(expandDescription("[shipdetails-groupmemberleader]"), column) + this.$padTextRight(this.$getGroupLeader(this._selectedItem.groupName), 32 - column) + "\n";
				}
				for (var i = 0; i < group.length; i++) {
					if (i === 0) {
						text += this.$padTextRight(expandDescription("[shipdetails-groupmembers]"), column);
					} else {
						text += this.$padTextRight("", column);
					}
					text += this.$padTextRight(group[i].shipType + (group[i].shipName.length > 1 ? ": " + group[i].shipName : ""), 32 - column) + "\n";
				}
			}
		}
		// ship escorts
		if (this._selectedItem.escortName !== "" && depart !== "Inbound") {
			if (this._selectedItem.groupName === "") text += "\n";
			var escort = this.$getEscortMembers(this._selectedItem.escortName, this._selectedItem);
			if (escort.length > 0) {
				if (this._selectedItem.escortLeader === true) {
					text += this.$padTextRight(expandDescription("[shipdetails-escortleader]"), column) + "Yes\n";
				} else {
					text += this.$padTextRight(expandDescription("[shipdetails-escortmemberleader]"), column) + this.$padTextRight(this.$getEscortLeader(this._selectedItem.escortName), 32 - column) + "\n";
				}
				for (var i = 0; i < escort.length; i++) {
					if (i === 0) {
						text += this.$padTextRight(expandDescription("[shipdetails-escortmembers]"), column);
					} else {
						text += this.$padTextRight("", column);
					}
					text += this.$padTextRight(escort[i].shipType + (escort[i].shipName.length > 1 ? ": " + escort[i].shipName : ""), 32 - column) + "\n";
				}
			}
		}

		// pilot info (if we have it)
		if (this._selectedItem.pilotName !== "") {
			text += "\n";
			text += this.$padTextRight(expandDescription("[shipdetails-pilotname]"), column);
			var colText = this.$columnText((this._selectedItem.isPlayer ? "Commander " : "") + this._selectedItem.pilotName + (this._selectedItem.pilotDescription !== "" ? ", " + this._selectedItem.pilotDescription : ""), 31 - column);
			for (var i = 0; i < colText.length; i++) {
				if (i !== 0) text += this.$padTextRight("", column);
				text += colText[i] + "\n";
			}
			if (this._selectedItem.pilotDescription === "") {
				if (this._selectedItem.pilotSpecies !== "") text += this.$padTextRight(expandDescription("[shipdetails-pilotspecies]"), column) + this.$toProperCase(this._selectedItem.pilotSpecies) + "\n";
				if (this._selectedItem.pilotHomeSystem !== -1) text += this.$padTextRight(expandDescription("[shipdetails-pilothomesystem]"), column) + System.systemNameForID(this._selectedItem.pilotHomeSystem) + "\n";
			}
		}

		if (this._selectedIndex === this._viewing.length - 1) {
			curChoices["90_NEXTSHIP"] = {
				text: "[option_nextship]",
				color: this._disabledColor,
				unselectable: true
			};
		} else {
			curChoices["90_NEXTSHIP"] = {
				text: "[option_nextship]",
				color: this._menuColor
			};
		}
		if (this._selectedIndex === 0) {
			curChoices["91_PREVSHIP"] = {
				text: "[option_prevship]",
				color: this._disabledColor,
				unselectable: true
			};
		} else {
			curChoices["91_PREVSHIP"] = {
				text: "[option_prevship]",
				color: this._menuColor
			};
		}
		if (this._showEquipment === true) curChoices["93_EQUIPMENT"] = {
			text: "[option_equipment]",
			color: this._menuColor
		};

		// only offer a launch before/after if the ship has a scheduled departure time visible
		if (p.dockedStation === this._selectedStation) {
			if (depart.indexOf(":") >= 0 && ((player.bounty === 0 && this._selectedItem.destinationHidden === false) || this._mode === 0)) {
				// and they're not due in the 10 minutes or less
				if (lt > (global.clock.adjustedSeconds + (10 * 60)) && depart !== "00:10") {
					curChoices["95_LAUNCH_BEFORE"] = {
						text: "[option_requestbefore]",
						color: this._menuColor
					};
				}
				if (lt >= (global.clock.adjustedSeconds + (10 * 60))) {
					curChoices["96_LAUNCH_AFTER"] = {
						text: "[option_requestafter]",
						color: this._menuColor
					};
				}
			}
		}
		if (hackchip && depart.indexOf(":") >= 0 && this._selectedItem.destinationHidden === true && lt >= (global.clock.adjustedSeconds + (10 * 60))) {
			curChoices["97_HACKCHIP"] = {
				text: "[hackchip-override]",
				color: this._menuColor
			};
		}

		// add the custom menu items here, if any
		for (var i = 0; i < this._externalMenu.length; i++) {
			// don't add menus to ships that are inbound
			if (depart !== "Inbound") {
				// is this menu limited to a specific station?
				var stnName = this._externalMenu[i].stationName;
				var stnIndex = parseInt(this._externalMenu[i].stationIndex);
				var stn = null;
				if (stnName !== "") stn = this._SDC.$getStationFromName(stnName, stnIndex);
				// if the specific station is not the current player station, just continue.
				if (stn !== null && stn !== player.ship.dockedStation) continue;
				// match based on shipname/type or pilot name
				if (this._externalMenu[i].shipName === "*" ||
					(this._externalMenu[i].shipName === this._selectedItem.shipName && this._externalMenu[i].shipType === this._selectedItem.shipType) ||
					(this._externalMenu[i].pilotName === this._selectedItem.pilotName)) {
					curChoices["80_EXT:" + (i < 10 ? "0" : "") + i.toString()] = {
						text: this._externalMenu[i].menuText,
						color: this._menuColor
					};
				}
			}
		}

		if (this._mode === 0) curChoices["97_LOG"] = {
			text: "[option_log]",
			color: this._menuColor
		};
		curChoices["98_CLOSE"] = {
			text: "[option_close]",
			color: this._exitColor
		};
		def = "98_CLOSE";
		var opts = {
			screenID: "oolite-sdc-ship-map",
			title: expandDescription("[ship_details_title]"),
			allowInterrupt: true,
			exitScreen: "GUI_SCREEN_INTERFACES",
			choices: curChoices,
			initialChoicesKey: this._shiplastchoice ? this._shiplastchoice : def,
			message: text,
		};
		if (this._shipModel === true) {
			opts["model"] = "[" + this._selectedItem.shipDataKey + "]";
			opts["modelPersonality"] = this._selectedItem.personality;
			opts["spinModel"] = this._spinModel;
		}
	}

	if (this._displayMode === 2) {
		var column = 11;

		// ship name/type
		text += ((this._selectedItem.shipName && this._selectedItem.shipName.length > 1) ? this.$padTextRight(expandDescription("[shipdetails-shipname]"), column) + this._selectedItem.shipName + "\n" : "") +
			this.$padTextRight(expandDescription("[shipdetails-shiptype]"), column) + this._selectedItem.shipType + ((this._mode === 0) ? " (" + this._selectedItem.primaryRole + ")" : "") + "\n\n";

		var eqlist = this.$equipmentListDisplay(this._selectedItem);
		if (eqlist.length > 0) {
			for (var i = 0; i < eqlist.length; i++) {
				text += this.$padTextRight((i === 0 ? "Equipment:" : ""), column) + eqlist[i] + "\n";
			}
		}
		curChoices["97_CLOSE"] = {
			text: "[option_close]",
			color: this._exitColor
		};
		def = "97_CLOSE";

		var opts = {
			screenID: "oolite-sdc-ship-map",
			title: expandDescription("[ship_equipment_title]"),
			allowInterrupt: true,
			exitScreen: "GUI_SCREEN_INTERFACES",
			choices: curChoices,
			initialChoicesKey: def,
			message: text
		};

		if (this._shipModel === true) {
			opts["model"] = "[" + this._selectedItem.shipDataKey + "]";
			opts["modelPersonality"] = this._selectedItem.personality;
			opts["spinModel"] = this._spinModel;
		}
	}

	mission.runScreen(opts, this.$screenHandler, this);

	var m = mission.displayModel;
	// rotate model for a better view
	if (m) {
		// if gallery has been loaded and used, remove the frame callback
		var w = worldScripts.gallery;
		if (w && isValidFrameCallback(w.$GalleryFCB)) removeFrameCallback(w.$GalleryFCB);

		var zoom_factor = this._defaultZoomFactor;
		if (this._displayMode === 1 && this._shipModel && this._spinModel === false) {
			if (this._zoomOverride[this._selectedItem.shipType]) zoom_factor = this._zoomOverride[this._selectedItem.shipType];
		}
		// zoom and rotate the image
		m.position = Vector3D(0, 0, m.collisionRadius * zoom_factor);
		m.orientation = m.orientation.rotate([0, 0, 0]);
		m.orientation = m.orientation.rotateZ(Math.PI * 1.2);
		m.orientation = m.orientation.rotateX(-1.1);
	}

	// refreshing the page would lose the XenonUI background, so force a reload here.
	if (!worldScripts.Lib_GUI) {
		if (worldScripts.XenonUI) worldScripts.XenonUI.guiScreenChanged("", "");
		if (worldScripts.XenonReduxUI) worldScripts.XenonReduxUI.guiScreenChanged("", "");
	} else {
		worldScripts.Lib_GUI.guiScreenChanged();
	}
}

//-------------------------------------------------------------------------------------------------------------
// handles UI responses on the dock list screen
this.$screenHandler = function $screenHandler(choice) {
	delete this._lastchoice;
	delete this._shiplastchoice;
	var newChoice = "";

	if (!choice) {
		return; // launched while reading?
	} else if (choice.indexOf("SHIPITEM") >= 0) {
		// get the selected item
		// use the index attached to the end of the choice
		this._selectedIndex = parseInt(choice.substring(choice.indexOf("SHIPITEM") + 9));
		this._selectedItem = this._viewing[this._selectedIndex];
		this._displayMode = 1;
		this._savedChoice = choice;
	} else if (choice.indexOf("80_EXT:") >= 0) {
		// call the external function
		var idx = parseInt(choice.substring(choice.indexOf(":") + 1));
		var w = worldScripts[this._externalMenu[idx].worldScript];
		var result = false;
		if (w) {
			result = w[this._externalMenu[idx].callback](this._selectedIndex, this._externalMenu[idx].parameter);
		}
		if (result === false) return;

	} else if (choice === "01_99_EXPAND") {
		this._showAll = true;
		newChoice = "10_GOTONEXT";
	} else if (choice === "11_GOTOPREV") {
		this._curpage -= 1;
		if (this._curpage === 0) {
			newChoice = "10_GOTONEXT";
		}
	} else if (choice === "10_GOTONEXT") {
		this._curpage += 1;
		if (this._curpage === this._maxpage - 1) {
			newChoice = "11_GOTOPREV";
		}
	} else if (choice === "31_SORTBY_SHIPNAME") {
		this._sortCol = "shipName";
		this._sortDir = 1;
		newChoice = "32_SORTBY_DEPART";
	} else if (choice === "30_SORTBY_SHIPTYPE") {
		this._sortCol = "shipType";
		this._sortDir = 1;
		newChoice = "31_SORTBY_SHIPNAME";
		if (this._rsnInstalled === false) newChoice = "32_SORTBY_DEPART";
	} else if (choice === "32_SORTBY_DEPART") {
		this._sortCol = "departureTime";
		this._sortDir = 1;
		newChoice = "31_SORTBY_SHIPNAME";
		if (this._rsnInstalled === false) newChoice = "30_SORTBY_SHIPTYPE";
	} else if (choice === "35_SORT_ASC") {
		this._sortDir = 1;
		newChoice = "36_SORT_DESC";
	} else if (choice === "36_SORT_DESC") {
		this._sortDir = -1;
		newChoice = "35_SORT_ASC";
	} else if (choice === "90_NEXTSHIP") {
		this._selectedIndex += 1;
		this._selectedItem = this._viewing[this._selectedIndex];
		this._shiplastchoice = choice;
		if (this._selectedIndex === this._viewing.length - 1) {
			this._shiplastchoice = "91_PREVSHIP";
		}
	} else if (choice === "91_PREVSHIP") {
		this._selectedIndex -= 1;
		this._selectedItem = this._viewing[this._selectedIndex];
		this._shiplastchoice = choice;
		if (this._selectedIndex === 0) {
			this._shiplastchoice = "90_NEXTSHIP";
		}
	} else if (choice === "93_EQUIPMENT") {
		this._displayMode = 2;

	} else if (choice === "95_LAUNCH_BEFORE") {
		// check if the ship will accept the request
		//if (this.$checkPermission(this._selectedItem) === false) {
		//	player.consoleMessage("Your request to launch before this ship was denied by the pilot.");
		// } else {
		// advance time to 11 minutes before selected ship departs
		var lt = (this._selectedItem.departureTime * 60 + this._selectedItem.lastChange);
		this.$advanceAndLaunch((lt - (10.5 * 60 + 30)) - global.clock.adjustedSeconds);
		choice = "99_EXIT";
		// }

	} else if (choice === "96_LAUNCH_AFTER") {
		// check if the ship will accept the request
		//if (this.$checkPermission(this._selectedItem) === false) {
		//	player.consoleMessage("Your request to launch after this ship was denied by the pilot.");
		// } else {
		// advance time to to 9 minutes before selected ship departs
		var lt = (this._selectedItem.departureTime * 60 + this._selectedItem.lastChange);
		this.$advanceAndLaunch((lt - (9 * 60 + 30)) - global.clock.adjustedSeconds);
		choice = "99_EXIT";
		// }
	} else if (choice === "97_HACKCHIP") {
		this._selectedItem.destinationHidden = false;
	} else if (choice === "97_LOG") {
		this._SDC.$writeDataToLog(this._selectedItem, "Logging data for ship record");
	} else if (choice === "97_CLOSE") {
		this._displayMode = 1;
	} else if (choice === "98_CLOSE") {
		this._displayMode = 0;
		this._selectedItem = {};
		newChoice = this._savedChoice;
		delete this._savedChoice;
	}

	this._lastchoice = choice;
	if (newChoice !== "") {
		this._lastchoice = newChoice;
	}

	if (choice !== "99_EXIT") {
		this.$showPage();
	}
}

//=============================================================================================================
// helper functions

//-------------------------------------------------------------------------------------------------------------
// function to return current state of repopulate control
this.$repopulateInProgress = function $repopulateInProgress() {
	return this._SDC._repopulate;
}

//-------------------------------------------------------------------------------------------------------------
// used by MFD to get list of ships in dock
this.$getStationDockList = function $getStationDockList(station) {
	function compare(a, b) {
		if ((a.departureTime * 10 + a.departureSeconds) < (b.departureTime * 10 + b.departureSeconds)) return -1;
		if ((a.departureTime * 10 + a.departureSeconds) > (b.departureTime * 10 + b.departureSeconds)) return 1;
		return 0;
	}

	if (this._SDC._repopulate === true) {
		this._viewing = [];
	} else {
		this._viewing = this._SDC.$getViewingList(station, true);
	}
	this._viewing.sort(compare);
	return this._viewing;
}

//-------------------------------------------------------------------------------------------------------------
// returns the shipname of the leader of a particular group
this.$getGroupLeader = function $getGroupLeader(groupName) {
	var leader = "";
	for (var i = 0; i < this._viewing.length; i++) {
		if (this._viewing[i].groupName === groupName && this._viewing[i].groupLeader === true) {
			leader = this._viewing[i].shipType + (this._viewing[i].shipName.length > 1 ? ": " + this._viewing[i].shipName : "");
			break;
		}
	}
	if (leader === "") leader = "(undocked vessel)";
	return leader;
}

//-------------------------------------------------------------------------------------------------------------
// returns the shipname of the leader of a particular escort
this.$getEscortLeader = function $getEscortLeader(escortName) {
	var leader = "";
	for (var i = 0; i < this._viewing.length; i++) {
		if (this._viewing[i].escortName === escortName && this._viewing[i].escortLeader === true) {
			leader = this._viewing[i].shipType + (this._viewing[i].shipName.length > 1 ? ": " + this._viewing[i].shipName : "");
			break;
		}
	}
	if (leader === "") leader = "(undocked vessel)";
	return leader;
}

//-------------------------------------------------------------------------------------------------------------
// returns the members of a particular group as a list of data records. "Exclude" is a record that will be left off
this.$getGroupMembers = function $getGroupMembers(groupName, exclude) {
	var shiplist = [];
	for (var i = 0; i < this._viewing.length; i++) {
		if (this._viewing[i].groupName === groupName && this._viewing[i].groupLeader === false && this._viewing[i] !== exclude) {
			shiplist.push(this._viewing[i]);
		}
	}
	return shiplist;
}

//-------------------------------------------------------------------------------------------------------------
// returns the members of a particular escort as a list of data records. "Exclude" is a record that will be left off
this.$getEscortMembers = function $getEscortMembers(escortName, exclude) {
	var shiplist = [];
	for (var i = 0; i < this._viewing.length; i++) {
		if (this._viewing[i].escortName === escortName && this._viewing[i].escortLeader === false && this._viewing[i] !== exclude) {
			shiplist.push(this._viewing[i]);
		}
	}
	return shiplist;
}

//-------------------------------------------------------------------------------------------------------------
// occasionally the shipUniqueName returns "" even though it's really there. In that instance, pull the unique name out of the display name
this.$getShipUniqueName = function $getShipUniqueName(displayName, shipClassName) {
	return displayName.replace(shipClassName + ": ", "");
}

//-------------------------------------------------------------------------------------------------------------
// return text as "Proper Case"
this.$toProperCase = function $toProperCase(text) {
	return text.replace(/\w\S*/g, function (txt) {
		return txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase();
	});
}

//-------------------------------------------------------------------------------------------------------------
// returns the leader item of a particular group
this.$getGroupLeaderItem = function $getGroupLeaderItem(groupName) {
	for (var i = 0; i < this._viewing.length; i++) {
		if (this._viewing[i].groupName === groupName && this._viewing[i].groupLeader === true) {
			return this._viewing[i];
		}
	}
	return null;
}

//-------------------------------------------------------------------------------------------------------------
// returns the leader item of a particular escort
this.$getEscortLeaderItem = function $getEscortLeaderItem(escortName) {
	for (var i = 0; i < this._viewing.length; i++) {
		if (this._viewing[i].escortName === escortName && this._viewing[i].escortLeader === true) {
			return this._viewing[i];
		}
	}
	return null;
}

//-------------------------------------------------------------------------------------------------------------
// appends space to currentText to the specified length in 'em'
this.$padTextRight = function $padTextRight(currentText, desiredLength) {
	if (currentText == null) currentText = "";
	var hairSpace = String.fromCharCode(31);
	var ellip = "…";
	var currentLength = defaultFont.measureString(currentText);
	var hairSpaceLength = defaultFont.measureString(hairSpace);
	// calculate number needed to fill remaining length
	var padsNeeded = Math.floor((desiredLength - currentLength) / hairSpaceLength);
	if (padsNeeded < 1) {
		// text is too long for column, so start pulling characters off
		var tmp = currentText;
		do {
			tmp = tmp.substring(0, tmp.length - 2) + ellip;
			if (tmp === ellip) break;
		} while (defaultFont.measureString(tmp) > desiredLength);
		currentLength = defaultFont.measureString(tmp);
		padsNeeded = Math.floor((desiredLength - currentLength) / hairSpaceLength);
		currentText = tmp;
	}
	// quick way of generating a repeated string of that number
	return currentText + new Array(padsNeeded).join(hairSpace);
}

//-------------------------------------------------------------------------------------------------------------
// arranges text into a array of strings with a particular column width
this.$columnText = function $columnText(originalText, columnWidth) {
	var returnText = [];
	if (defaultFont.measureString(originalText) > columnWidth) {
		var hold = originalText;
		do {
			var newline = "";
			var remain = "";
			var point = hold.length;
			do {
				point = hold.lastIndexOf(" ", point - 1);
				newline = hold.substring(0, point).trim();
				remain = hold.substring(point + 1).trim();
			} while (defaultFont.measureString(newline) > columnWidth);
			returnText.push(newline);
			if (remain !== "") {
				if (defaultFont.measureString(remain) <= columnWidth) {
					returnText.push(remain);
					hold = "";
				} else {
					hold = remain;
				}
			} else {
				hold = "";
			}
		} while (hold !== "");
	} else {
		returnText.push(originalText);
	}
	return returnText;
}

//-------------------------------------------------------------------------------------------------------------
// duplicates text until it is just less than the desired length;
this.$duplicate = function $duplicate(text, desiredLength) {
	var res = "";

	do {
		res += text;
	} while (defaultFont.measureString(res) < desiredLength);

	res = res.substring(1, res.length);

	return res;
}

//-------------------------------------------------------------------------------------------------------------
// returns the docking bay prefix for the selected station
this.$getDockingBayPrefix = function $getDockingBayPrefix(station) {
	if (!station) return "";
	var pref = station.name.toString().substring(0, 1) + String.fromCharCode(65 + (system.ID % 26)) + "-";

	if (station.mass < 50000000) {
		// some smaller stations won't have a prefix
		switch (system.ID % 5) {
			case 0:
			case 2:
				pref = "";
				break;
		}
	} else {
		switch (system.ID % 5) {
			case 0:
				pref += "P-";
				break;
			case 1:
				pref += "T-";
				break;
			case 2:
				pref += "S-";
				break;
			case 3:
				pref += "L-";
				break;
			case 4:
				pref += "G-";
				break;
		}
	}

	return pref;
}

//-------------------------------------------------------------------------------------------------------------
// adds the player to the stations docked list
this.$addPlayerToArray = function $addPlayerToArray(stationList) {
	var p = player.ship;
	if (this._playerDockingBay === "") {
		this._playerDockingBay = this._SDC.$getDockingBayCode(p.dockedStation);
	}

	stationList.push({
		system: system.ID,
		station: p.dockedStation.name,
		shipName: p.shipUniqueName,
		shipType: p.shipClassName,
		shipDataKey: p.dataKey,
		personality: p.entityPersonality,
		primaryRole: p.primaryRole,
		shipAI: "",
		accuracy: 0,
		equipment: "",
		heatInsulation: 0,
		homeSystem: -1,
		bounty: p.bounty,
		escortName: "",
		escortLeader: false,
		groupName: "",
		groupLeader: false,
		destinationSystem: p.targetSystem,
		destinationHidden: false,
		goods: "",
		pilotDescription: "",
		pilotName: player.name,
		pilotHomeSystem: -1,
		pilotLegalStatus: p.bounty,
		pilotInsurance: 0,
		pilotSpecies: "",
		isPlayer: true,
		dockTime: this._playerDockTime,
		dockingBay: this._playerDockingBay,
		departureTime: (20 * 60),
		departureSeconds: 0
	});
}

//-------------------------------------------------------------------------------------------------------------
// returns a formatted version of a departure time (hh:mm)
this.$departureTimeFormat = function $departureTimeFormat(depart) {
	// get hours, minutes from depart
	//var sec_num = depart - global.clock.adjustedSeconds;

	if (depart === -10) {
		return "Inbound";
	} else if (Math.round(depart) === -5) {
		return "Rescheduled";
	} else if (depart < 1) {
		return "Launching";
	} else {
		var hours = Math.floor(depart / 60);
		var minutes = Math.floor(depart - (hours * 60));

		// determine status based on time range
		if (hours >= 23) return "Docking";
		if (hours >= 20 && hours < 23) return "Unloading";
		if (hours >= 16 && hours < 20) return "Docked";
		if (hours >= 12 && hours < 16) return "Loading";

		return (hours < 10 ? "0" : "") + hours + ":" + (minutes < 10 ? "0" : "") + minutes;
	}
}

//-------------------------------------------------------------------------------------------------------------
// reurns false if station has a role in the list of roles where no view is allowed. Otherwise true
this.$checkStationRoleForView = function $checkStationRoleForView(station) {
	for (var i = 0; i < station.roles.length; i++) {
		if (this._noDockView.indexOf(station.roles[i]) >= 0) return false;
	}
	return true;
}

//-------------------------------------------------------------------------------------------------------------
// interprets the data of a particular entry and determines whether the destination is viewable or not
this.$viewDestination = function $viewDestination(item) {
	// if the destination is set (not equal to -1)
	//  and player.bounty equals 0
	// 	and not equal to the current system.ID,
	//  and the destinationHidden flag is not set,
	//  and the departure time is within the viewable period
	// then return true (ie the destination is viewable), otherwise false
	var lt = (item.departureTime * 60 + item.lastChange)
	if (item.destinationSystem !== -1 && player.bounty === 0 &&
		(item.destinationSystem !== system.ID || (item.destination && item.destination !== "")) &&
		item.destinationHidden === false &&
		lt < (global.clock.adjustedSeconds + (this._destViewingRange * 60 * 60))) return true;
	return false;
}

//-------------------------------------------------------------------------------------------------------------
this.$viewLocation = function $viewLocation(item) {
	if (item) {
		var items = item.split("|");
		return items[0];
	} else {
		return "";
	}
}

//-------------------------------------------------------------------------------------------------------------
this.$outputDestination = function $outputDestination(item) {
	if (item.destination && item.destination !== "") {
		return this.$viewLocation(item.destination);
	} else {
		return System.systemNameForID(item.destinationSystem);
	}
}

//-------------------------------------------------------------------------------------------------------------
// work out if the ship will allow the player to launch with them
// returns true if allowed, otherwise false
this.$checkPermission = function $checkPermission(item) {
	var retVal = true;
	var checking = item;
	// get the parent item if this is an escort
	if (item.escortName !== "" && item.escortLeader === false) {
		checking = this.$getEscortLeaderItem(item.escortName);
	}
	if (item.groupName !== "" && item.groupLeader === false) {
		checking = this.$getGroupLeaderItem(item.groupName);
	}
	var checkrole = player.roleWeights[parseInt(system.scrambledPseudoRandomNumber(checking.personality) * player.roleWeights.length)];
	switch (checking.primaryRole) {
		case "shuttle":
		case "trader":
		case "trader-courier":
		case "trader-smuggler":
			if (checkrole.indexOf("pirate") >= 0 || checkrole.indexOf("assassin") >= 0) retVal = false;
			break;
		case "pirate":
		case "pirate-light-freighter":
		case "pirate-medium-freighter":
			if (checkrole.indexOf("hunter") >= 0 || checkrole.indexOf("assassin") >= 0) retVal = false;
			break;
		case "assassin-light":
		case "assassin-medium":
		case "assassin-heavy":
			if (checkrole.indexOf("assassin") >= 0) retVal = false;
			break;
		case "hunter":
		case "hunter-medium":
		case "hunter-heavy":
			if (checkrole.indexOf("pirate") >= 0 || checkrole.indexOf("assassin") >= 0) retVal = false;
			break;
	}
	if (this._debug) log(this.name, "Checking permission for " + checking.primaryRole + " against player role " + checkrole + ": result " + retVal);
	return retVal;
}

//-------------------------------------------------------------------------------------------------------------
// advance a given amount of time and launch the player
this.$advanceAndLaunch = function $advanceAndLaunch(advanceTime) {

	var p = player.ship;
	if (this._debug) log(this.name, "Advancing time " + advanceTime + " seconds");
	global.clock.addSeconds(advanceTime);
	var sdc = worldScripts.StationDockControl;
	if (sdc && sdc._disable === false) sdc.$checkForLaunchedShips();

	// launch player
	if (this._debug) log(this.name, "Launching the player");
	if (this.$isBigGuiActive() === false) {
		p.hudHidden = false;
	}
	p.launch();
	this._hudHidden = false;
}

//-------------------------------------------------------------------------------------------------------------
// returns true if a HUD with allowBigGUI is enabled, otherwise false
this.$isBigGuiActive = function $isBigGuiActive() {
	if (oolite.compareVersion("1.83") <= 0) {
		return player.ship.hudAllowsBigGui;
	} else {
		var bigGuiHUD = ["XenonHUD.plist", "coluber_hud_ch01-dock.plist"]; // until there is a property we can check, I'll be listing HUD's that have the allow_big_gui property set here
		if (bigGuiHUD.indexOf(player.ship.hud) >= 0) {
			return true;
		} else {
			return false;
		}
	}
}

//-------------------------------------------------------------------------------------------------------------
// returns array of equipment item descriptions for the specified ship item
this.$equipmentListDisplay = function $equipmentListDisplay(item) {
	var eqList = [];
	if (item.equipment !== "") {
		var list = item.equipment.split(",");
		for (var i = 0; i < list.length; i++) {
			var eq = list[i];
			var def = "";
			var extra = "";
			if (list[i].indexOf(":") >= 0) {
				var subs = list[i].split(":");
				def = subs[0];
				eq = subs[1];
			}
			if (def !== "X") {
				switch (def) {
					case "FORE":
						extra = "Forward laser: ";
						break;
					case "AFT":
						extra = "Aft laser: ";
						break;
					case "PORT":
						extra = "Port laser: ";
						break;
					case "STARBOARD":
						extra = "Starboard laser: ";
						break;
				}
			}
			if (eq !== "") {
				var eqDef = EquipmentInfo.infoForKey(eq);
				if (eqDef) eqList.push(extra + eqDef.name);
			}
		}
	} else {
		var ship = Ship.shipDataForKey(item.shipDataKey);
		var weapon = this.$translateWeaponInfo(ship.forward_weapon_type);
		if (weapon !== "") eqList.push("Forward laser: " + weapon);
		weapon = this.$translateWeaponInfo(ship.aft_weapon_type);
		if (weapon !== "") eqList.push("Aft laser: " + weapon);
		weapon = this.$translateWeaponInfo(ship.port_weapon_type);
		if (weapon !== "") eqList.push("Port laser: " + weapon);
		weapon = this.$translateWeaponInfo(ship.starboard_weapon_type);
		if (weapon !== "") eqList.push("Starboard laser: " + weapon);
		if (this.$checkValue(ship.has_cloaking_device) === true) eqList.push("Cloaking Device");
		if (this.$checkValue(ship.has_ecm) === true) eqList.push("E.C.M. System");
		if (this.$checkValue(ship.has_escape_pod) === true) eqList.push("Escape Pod");
		if (this.$checkValue(ship.has_fuel_injection) === true) eqList.push("Witchdrive Fuel Injectors");
		if (this.$checkValue(ship.has_scoop) === true) eqList.push("Fuel Scoops");
		if (this.$checkValue(ship.has_shield_booster) === true) eqList.push("Shield Boosters");
		if (this.$checkValue(ship.has_shield_enhancer) === true) eqList.push("Military Shield Enhancement");
	}
	return eqList;
}

//-------------------------------------------------------------------------------------------------------------
// translates the weapon info into an actual name
this.$translateWeaponInfo = function $translateWeaponInfo(dta) {
	var result = "";
	if (!dta == null && dta !== "WEAPON_NONE") {
		var wpn = "EQ_" + dta;
		var eq = EquipmentInfo.infoForKey(wpn);
		result = eq.name;
	}
	return result;
}

//-------------------------------------------------------------------------------------------------------------
// checks a value for true or false
this.$checkValue = function $checkValue(dta) {
	if (!dta == null) {
		if (dta === "yes" || dta === "true" || dta === "1" || parseInt(dta) > 1) {
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}
